import arcpy
import os, csv, datetime, sys, subprocess, numpy

"""
/***************************************************************************
 ArcDayMet
                             -------------------
        begin                : 2016-08-29
        copyright            : (C) 2016 by Apropos Information Systems Inc.
        email                : info@aproposinfosystems.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   any later version.                                                    *
 *                                                                         *
 ***************************************************************************/
"""

class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "ArcDayMet Toolbox (1.0.1)"
        self.alias = "arcdaymet"

        # List of tool classes associated with this toolbox
        self.tools = [DayMetExtract]

#
# DayMet Exraction Tool
#
class DayMetExtract(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Extract DayMet Data"
        self.description = "Extract and summarize DayMet data for points in shapefile."
        self.canRunInBackground = True
        # variables
        self.fullNames = ["Maximum Temperature","Minimum Temperature","Shortwave Radiation","Vapor Pressure","Snow-water Equivalent","Precipitation","Day Length"]
        self.shortNames = ["tmax","tmin","srad","vp","swe","prcp","dayl"]
        self.aggregateVariables = ["Mean","Median","Variance","Standard Deviation","Count","Sum","Maximum","Minimum"]

    def getParameterInfo(self):
        """Define parameter definitions"""
        # point layer 
        point_layer = arcpy.Parameter(
            displayName="Point Layer (locations to extract DayMet data)",
            name="point_layer",
            datatype="GPFeatureLayer",
            parameterType="Required",
            direction="Input")
        point_layer.filter.list = ["Point"]
        
        # point id field name 
        pid_field = arcpy.Parameter(
            displayName="Point Id Field (a unique identifier for each location)",
            name="pid_field",
            datatype="Field",
            parameterType="Required",
            direction="Input")
        pid_field.parameterDependencies = [point_layer.name]

        # daymet variables
        daymet_variables = arcpy.Parameter(
            displayName="DayMet Variables (select one or more to record for each point)",
            name="daymet_variables",
            datatype="GPString",
            parameterType="Required",
            direction="Input",
            multiValue=True)
        daymet_variables.filter.list = self.fullNames
        
        minYear = 1980
        maxYear = datetime.datetime.now().year - 1
        # start year 
        start_year = arcpy.Parameter(
            displayName="Start Year (first year to extract DayMet variables)",
            name="start_year",
            datatype="GPLong",
            parameterType="Required",
            direction="Input")
        start_year.filter.type = "Range"
        start_year.filter.list = [minYear,maxYear]
        
        # end year
        end_year = arcpy.Parameter(
            displayName="End Year (last year to extract DayMet variables)",
            name="end_year",
            datatype="GPLong",
            parameterType="Required",
            direction="Input")
        end_year.filter.type = "Range"
        end_year.filter.list = [minYear,maxYear]

        # data processing
        data_processing = arcpy.Parameter(
            displayName="Data Processing (how to process raw data)",
            name="data_processing",
            datatype="GPString",
            parameterType="Required",
            direction="Input")
        data_processing.filter.list = ["Return Raw Data","Return Raw Data and Aggregates"]
        data_processing.value = "Return Raw Data"
        
        # group field name 
        group_field = arcpy.Parameter(
            displayName="Group Field (a field to aggregate data)",
            name="group_field",
            datatype="Field",
            parameterType="Optional",
            direction="Input")
        group_field.parameterDependencies = [point_layer.name]
        group_field.enabled = False

        # aggregate variables
        aggregate_variables = arcpy.Parameter(
            displayName="Aggregate Measures (select what aggregate calculations you want)",
            name="aggregate_variables",
            datatype="GPString",
            parameterType="Optional",
            direction="Input",
            multiValue=True)
        aggregate_variables.filter.list = self.aggregateVariables
        #aggregate_variables.value = ["Mean","Variance","Count","Sum"]
        aggregate_variables.enabled = False
        
        # output file
        output_file = arcpy.Parameter(
            displayName="Output File (csv of results)",
            name="output_file",
            datatype="DEFile",
            parameterType="Required",
            direction="Output")
        output_file.filter.list = ['csv']
        
        params = [point_layer, pid_field, daymet_variables, start_year, end_year, data_processing, group_field, aggregate_variables, output_file]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        data_processing = parameters[5]
        group_field = parameters[6]
        aggregate_variables = parameters[7]
        if data_processing.valueAsText == "Return Raw Data":
            group_field.enabled = False
            group_field.parameterType = 'Optional'
            aggregate_variables.enabled = False
            aggregate_variables.parameterType = 'Optional'
        else:
            group_field.enabled = True
            group_field.parameterType = 'Required'
            aggregate_variables.enabled = True
            aggregate_variables.parameterType = 'Required'
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        # give parameters meaningful names
        sYear = parameters[3]
        eYear = parameters[4]
        data_processing = parameters[5]
        group_field = parameters[6]
        aggregate_variables = parameters[7]
        # check that start year is before end year 
        if not sYear.value is None:
            if not eYear.value is None:
                if sYear.value > eYear.value:
                    sYear.setErrorMessage("Start Year must be same or before End Year")
                    eYear.setErrorMessage("End Year must be same or after Start Year")
            else:
                eYear.setErrorMessage("End Year must be set")
        else:
            sYear.setErrorMessage("Start Year must be set")
        # ensure group id set if returning aggregate values
        if (group_field.valueAsText == "" or group_field.valueAsText is None) and data_processing.valueAsText == "Return Raw Data and Aggregates":
            group_field.setErrorMessage("Group Field must be selected")
        # ensure aggregate variables selected if aggregate returns selected
        if aggregate_variables.valueAsText is None and data_processing.valueAsText == "Return Raw Data and Aggregates":
            aggregate_variables.setErrorMessage("Aggregate Measures must be selected")
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        self.debugMode = False
        layer = parameters[0].valueAsText
        idField = parameters[1].valueAsText
        # because some parameters are single words and others double words, need to get rid of single quotes before split so no extra quotes added
        dayMetVariables = parameters[2].valueAsText.replace("'",'').split(';')
        startYear = parameters[3].value
        endYear = parameters[4].value
        dataProcessing = parameters[5].valueAsText
        groupField = parameters[6].valueAsText
        # because some parameters are single words and others double words, need to get rid of single quotes before split so no extra quotes added
        if parameters[7].valueAsText is None:
            aggregateVariables = []
        else:
            aggregateVariables = parameters[7].valueAsText.replace("'",'').split(';')
        outFName = parameters[8].valueAsText
        oDir,oFile = os.path.split(outFName)
        dirname,fname = os.path.split(os.path.abspath(__file__))
        self.execFile = os.path.join(dirname,'curl.exe')
        if not os.path.exists(self.execFile):
            arcpy.AddError('curl.exe was not found. It must be in the same folder as ArcDayMet.pyt')
        elif os.path.exists(outFName):
            arcpy.AddError('Can not overwrite file: %s' % outFName)
        else:
            # create param list string
            pString = ''
            for p in dayMetVariables:
                if p in self.fullNames:
                    pString = pString + self.shortNames[self.fullNames.index(p)] + ','
            pString = pString[:-1]
            # create year list string
            yString = ",".join([str(i) for i in range(startYear,endYear+1)])
            # create download string
            base_string = "https://daymet.ornl.gov/single-pixel/send/saveData?lat=my_lat&lon=my_lon&measuredParams=param_list&year=year_range&format=csv"            
            # substitute input variables in download string
            download_string = base_string.replace("param_list",pString)
            download_string = download_string.replace("year_range",yString)
            if self.debugMode == True:
                # check inputs
                pString = pString[:-1]
                pf = open(os.path.join(oDir,'params.txt'),'w')
                pf.write('layer: %s\n' % layer)
                pf.write('idField: %s\n' % idField)
                pf.write('variables: %s\n' % str(dayMetVariables))
                pf.write('pString: %s\n' % pString)
                pf.write('startYear: %s\n' % startYear)
                pf.write('endYear: %s\n' % endYear)
                pf.write('outFName: %s\n' % outFName)
                pf.write('base_string: %s\n' % base_string)
                pf.write('download_string: %s\n' % download_string)
                pf.close()
            # now process each point in layer
            fileList = self.processPoints(messages,layer,idField,groupField,download_string,oDir)
            # unify files into a single csv
            self.mergeFiles(messages,fileList,outFName)
            # aggregate if needed
            if dataProcessing == "Return Raw Data and Aggregates":
                self.aggregateData(messages,outFName,aggregateVariables)
        
        return

    def processPoints(self, messages, layer, idField, groupField, download_string, oDir):
    
        fileList = []
        if groupField is None:
            fields = ["SHAPE@XY",idField]
        else:
            fields = ["SHAPE@XY",idField,groupField]
        # hide terminal window for each execution of curl
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        # get projection information
        spatialRef = arcpy.Describe(layer).spatialReference
        if self.debugMode == True:
            cf = open(os.path.join(oDir,'commands.txt'),'w')
        if spatialRef <> arcpy.SpatialReference(4326):
            transformNeeded = True
        else:
            transformNeeded = False
        for row in arcpy.da.SearchCursor(layer,fields):
            if transformNeeded:
                x,y = row[0]
                point = arcpy.Point()
                point.X = x
                point.Y = y 
                pointGeom = arcpy.PointGeometry(point,spatialRef)
                nPointGeom = pointGeom.projectAs(arcpy.SpatialReference(4326))
                nCentroid = nPointGeom.centroid
                newX = nCentroid.X 
                newY = nCentroid.Y
            else:
                newX,newY = row[0]
            pid = row[1]
            if groupField is None:
                gid = 0
            else:
                gid = row[2]
            call_string = download_string.replace("my_lat",str(newY))
            call_string = call_string.replace("my_lon",str(newX))
            if self.debugMode == True:
                cf.write(call_string+'\n')
            # create filename for the output file
            daymet_file = os.path.join(oDir,'daymet-temp-%s.csv'% str(pid)) 
            # download the daymet data (if available)
            time.sleep(2)
            cmd = [self.execFile,'-X','GET','-o',daymet_file,call_string,'-k','-L'] # look at this
            process = subprocess.Popen(cmd,stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE,bufsize=4096,startupinfo=startupinfo)
            process.communicate()
            df = open(daymet_file,'r')
            outsideRange = False
            # check file if it is small
            if os.path.getsize(daymet_file) < 5000:
                for line in df:
                    if 'HTTP Status 500' in line:
                        outsideRange = True
                        break
                df.close()
            if outsideRange == True:
                os.remove(daymet_file)
            else:
                fileList.append([gid,pid,newX,newY,daymet_file])
        fileList.sort()
        if self.debugMode == True:
            cf.close()
        return(fileList)
        
    def mergeFiles(self, messages, fileList, outFName):
        
        ofn = open(outFName,'w')
        headerWritten = False
        for rec in fileList:
            f = open(rec[4],'r')
            dataStarted = False
            for line in f.readlines():
                if not dataStarted:
                    if 'year,yday' in line:
                        dataStarted = True
                        if headerWritten == False:
                            ofn.write('pointid,groupid,x,y,'+line)
                            headerWritten = True
                else:
                    outLine = '%s,%s,%f,%f,%s' % (str(rec[1]),str(rec[0]),rec[2],rec[3],line)
                    ofn.write(outLine)
            f.close()
            os.remove(rec[4])
        ofn.close()
        
    def aggregateData(self, messages, outFName, aggregateVariables):
        #
        # input format is as follows:
        # pointid, groupid, x, y, year, yday, <data fields>
        #
        # this aggregation funcation aggregated data by year and day across multiple points
        #
        aggFName,ext = os.path.splitext(outFName)
        aggFName = aggFName + '_aggregated' + ext
        inF = open(outFName,'r')
        outF = open(aggFName,'w')
        groupKey = None
        groups = {}
        # assemble raw data
        for line in inF:
            vals = line.split(',')
            if vals[0] == 'pointid':
                # first line of merged csv file
                # read header and write appropriate header to aggregate document
                headString = 'groupid,year,yday,'
                for val in vals[6:]:
                    sn = val.split(' ')[0]
                    if "Mean" in aggregateVariables:
                        headString = headString + sn + '_mean,'
                    if "Median" in aggregateVariables:
                        headString = headString + sn + '_median,'
                    if "Variance" in aggregateVariables:
                        headString = headString + sn + '_variance,'
                    if "Standard Deviation" in aggregateVariables:
                        headString = headString + sn + '_standard_deviation,'
                    if "Count" in aggregateVariables:
                        headString = headString + sn + '_count,'
                    if "Sum" in aggregateVariables:
                        headString = headString + sn + '_sum,'
                    if "Maximum" in aggregateVariables:
                        headString = headString + sn + '_max,'
                    if "Minimum" in aggregateVariables:
                        headString = headString + sn + '_min,'
                outF.write(headString[:-1] + '\n')
            else:
                # create key made up of groupid, year and day of year
                rowKey = str(vals[1]) + '|||' + str(vals[4]) + '|||' + str(vals[5])
                # check if key is already in dictionary
                if rowKey in groups:
                    groups[rowKey][1].append([float(z) for z in vals[6:]])
                else:
                    groups[rowKey] = [[vals[1],int(vals[4]),int(vals[5])],[[float(z) for z in vals[6:]]]]
        # aggregate data
        groupList = []
        for key,value in groups.iteritems():
            groupMeans = list(numpy.mean(numpy.array(value[1]),axis=0))
            groupMedians = list(numpy.median(numpy.array(value[1]),axis=0))
            groupVariances = list(numpy.var(numpy.array(value[1]),axis=0))
            groupStandardDeviations = list(numpy.std(numpy.array(value[1]),axis=0))
            groupCounts = len(value[1])
            groupSums = list(numpy.sum(numpy.array(value[1]),axis=0))
            groupMaximums = list(numpy.nanmax(numpy.array(value[1]),axis=0))
            groupMinimums = list(numpy.nanmin(numpy.array(value[1]),axis=0))
            tempList = []
            for x in range(len(groupMeans)):
                if "Mean" in aggregateVariables:
                    tempList.append(groupMeans[x])
                if "Median" in aggregateVariables:
                    tempList.append(groupMedians[x])
                if "Variance" in aggregateVariables:
                    tempList.append(groupVariances[x])
                if "Standard Deviation" in aggregateVariables:
                    tempList.append(groupStandardDeviations[x])
                if "Count" in aggregateVariables:
                    tempList.append(groupCounts)
                if "Sum" in aggregateVariables:
                    tempList.append(groupSums[x])
                if "Maximum" in aggregateVariables:
                    tempList.append(groupMaximums[x])
                if "Minimum" in aggregateVariables:
                    tempList.append(groupMinimums[x])
            groupList.append([value[0],tempList])
        groupList.sort()
        for rec in groupList:
            outLine = ','.join(str(e) for e in rec[0]) + ',' + ','.join(str(e) for e in rec[1])
            outF.write(outLine + '\n')
        inF.close()
        outF.close()
